from abc import ABC, abstractmethod
import os
from typing import TYPE_CHECKING, Dict, List, Union

from omegaconf import DictConfig

from autrainer.core.constants import ExportConstants
from autrainer.core.utils import Timer
from autrainer.metrics import AbstractMetric


if TYPE_CHECKING:
    from autrainer.training import ModularTaskTrainer  # pragma: no cover


def get_params_to_export(
    params: Union[dict, DictConfig], prefix: str = ""
) -> Dict[str, Union[int, float, str]]:
    """Get parameters to export as a flattened dictionary filtered by
    :const:`~autrainer.core.constants.ExportConstants.IGNORE_PARAMS`
    and :const:`~autrainer.core.constants.ExportConstants.LOGGING_DEPTH`.

    Private parameters (starting with "_") are also ignored except for
    "_target_".

    Args:
        params: The parameters of the configuration.
        prefix: The prefix to add to the keys. Defaults to "".

    Returns:
        The filtered and flattened dictionary of parameters.
    """
    result = {}
    for k, v in params.items():
        full_key = f"{prefix}.{k}" if prefix else k
        if (
            full_key in ExportConstants().IGNORE_PARAMS
            or prefix in ExportConstants().IGNORE_PARAMS
            or len(full_key.split(".")) > ExportConstants().LOGGING_DEPTH
            or prefix.startswith("_")
            or (k.startswith("_") and k != "_target_")
        ):
            continue
        if isinstance(v, (dict, DictConfig)):
            cfg_id = v.get("id", None)
            if cfg_id is None and v.get("_target_", None):
                raise KeyError(
                    f"Configuration '{full_key}' is missing an id field."
                )
            if cfg_id == "None" or not v.keys():  # placeholder syntax
                continue
            elif cfg_id is None:  # shorthand syntax
                result[full_key] = next(iter(v.keys()))
            else:  # full syntax
                result[full_key] = v.pop("id")
                result.update(get_params_to_export(v, prefix=full_key))
        else:
            result[full_key] = v
    return result


class AbstractLogger(ABC):
    def __init__(
        self,
        exp_name: str,
        run_name: str,
        metrics: List[AbstractMetric],
        tracking_metric: AbstractMetric,
        artifacts: List[
            Union[str, Dict[str, str]]
        ] = ExportConstants().ARTIFACTS,
    ) -> None:
        """Base class for loggers.

        Args:
            exp_name: The name of the experiment.
            run_name: The name of the run.
            metrics: The metrics to log.
            tracking_metric: The metric to determine the best results.
            artifacts: The artifacts to log. Defaults to
                :const:`~autrainer.core.constants.ExportConstants.ARTIFACTS`.
        """
        self.run_name = run_name
        self.exp_name = exp_name
        self.metrics = metrics
        self.tracking_metric = tracking_metric
        self.artifacts = artifacts
        self.best_metrics = {
            "train_loss.min": float("inf"),
            "dev_loss.min": float("inf"),
            "best_iteration": 0,
        }
        self.metrics_dict: Dict[str, AbstractMetric] = {}
        for metric in self.metrics:
            self.best_metrics[f"{metric.name}.{metric.suffix}"] = (
                metric.starting_metric
            )
            self.metrics_dict[metric.name] = metric

    def log_and_update_metrics(
        self, metrics: Dict[str, Union[int, float]], iteration: int = None
    ) -> None:
        """Log all metrics in the metrics dictionary for the given iteration.
        Automatically updates the best metrics based on the tracking metric.

        Args:
            metrics: The metrics to log, with the metric name as the key and
                the metric value as the value.
            iteration: The iteration, epoch, or step number.
                If None, the iteration will not be logged
                (e.g., for test metrics). Defaults to None.
        """
        if iteration:
            self._update_best_metrics(metrics, iteration)
            self.log_metrics(self.best_metrics)
        self.log_metrics(metrics, iteration)

    def _update_best_metrics(
        self,
        metrics: Dict[str, Union[int, float]],
        iteration: int,
    ) -> None:
        if self.tracking_metric.compare(
            metrics[self.tracking_metric.name],
            self.best_metrics[
                f"{self.tracking_metric.name}.{self.tracking_metric.suffix}"
            ],
        ):
            self.best_metrics["best_iteration"] = iteration
        for k, v in metrics.items():
            if ".std" in k:
                continue
            if "loss" in k:
                if v < self.best_metrics[k + ".min"]:
                    self.best_metrics[k + ".min"] = v
            else:
                metric = self.metrics_dict[k]
                if metric.compare(
                    v, self.best_metrics[f"{k}.{metric.suffix}"]
                ):
                    self.best_metrics[f"{k}.{metric.suffix}"] = v

    def setup(self) -> None:
        """Optional setup method called at the beginning of the run
        (`cb_on_train_begin`).
        """

    @abstractmethod
    def log_params(self, params: Union[dict, DictConfig]) -> None:
        """Log the parameters of the configuration.

        Args:
            params: The parameters of the configuration.
        """

    @abstractmethod
    def log_metrics(
        self,
        metrics: Dict[str, Union[int, float]],
        iteration=None,
    ) -> None:
        """Log the metrics for the given iteration.

        Args:
            metrics: The metrics to log, with the metric name as the key and
                the metric value as the value.
            iteration: The iteration, epoch, or step number.
                If None, the iteration will not be logged
                (e.g., for test metrics). Defaults to None.
        """

    @abstractmethod
    def log_timers(self, timers: Dict[str, float]) -> None:
        """Log all timers (e.g., mean times for train, dev, and test).

        Args:
            timers: The timers to log, with the timer name as the key and the
                timer value as the value.
        """

    @abstractmethod
    def log_artifact(self, filename: str, path: str = "") -> None:
        """Log an artifact (e.g., a file).

        Args:
            filename: The name of the artifact.
            path: The absolute path or relative path from the current working
                directory to the artifact. Defaults to "".
        """

    def end_run(self) -> None:
        """Optional end run method called at the end of the run
        (`cb_on_train_end`).
        """

    def cb_on_train_begin(self, trainer: "ModularTaskTrainer") -> None:
        self.setup()
        self.log_params(trainer.cfg)

    def cb_on_iteration_end(
        self, trainer: "ModularTaskTrainer", iteration: int, metrics: dict
    ) -> None:
        self.log_and_update_metrics(metrics, iteration)

    def cb_on_test_end(
        self, trainer: "ModularTaskTrainer", test_results: dict
    ) -> None:
        self.log_and_update_metrics(test_results)

    def cb_on_train_end(self, trainer: "ModularTaskTrainer") -> None:
        self.log_timers(
            {
                "time." + t.timer_type + ".mean": Timer.pretty_time(
                    t.get_mean_seconds()
                )
                for t in [
                    trainer.train_timer,
                    trainer.dev_timer,
                    trainer.test_timer,
                ]
            }
        )
        for artifact in self.artifacts:
            if isinstance(artifact, dict):
                for filename, path in artifact.items():
                    self.log_artifact(
                        filename, os.path.join(trainer.output_directory, path)
                    )
            else:
                self.log_artifact(artifact, trainer.output_directory)

        self.end_run()



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, List, Mapping, Type, TypeVar, Union

import numpy as np
import torch


AbstractItemType = TypeVar("ItemType", bound="AbstractDataItem")


@dataclass
class AbstractDataItem(ABC):
    """Abstract data item class for a single sample.

    Args:
        features: Tensor of input features.
        target: Target value for the input features.
        index: Index of the data sample.
    """

    features: torch.Tensor
    target: Union[int, float, List[int], List[float], np.ndarray]
    index: int


@dataclass
class AbstractDataBatch(ABC, Generic[AbstractItemType]):
    """Abstract data batch class for a batch of samples.

    Args:
        features: Tensor of input features.
        target: Tensor of target values for the input features.
        index: Tensor of indices for the data samples.
    """

    features: torch.Tensor
    target: torch.Tensor
    index: torch.Tensor

    @abstractmethod
    def to(self, device: torch.device, **kwargs: dict) -> None:
        """Move the features, target, and additional data to a device.

        Args:
            device: Device to move the data to.
            kwargs: Additional keyword arguments passed to the `to` method.
        """

    @classmethod
    def collate(
        cls: Type["AbstractDataBatch[AbstractItemType]"],
        items: List[AbstractItemType],
    ) -> "AbstractDataBatch[AbstractItemType]":
        """Collate a list of data items into a data batch.

        Args:
            items: List of data items to collate.

        Returns:
            Collated data batch.
        """
        if not isinstance(items[0], Mapping):
            items = [vars(item) for item in items]
        first = items[0]
        batch = {}

        for k, v in first.items():
            if isinstance(v, torch.Tensor):
                batch[k] = torch.stack([f[k] for f in items])
            elif isinstance(v, np.ndarray):
                batch[k] = torch.from_numpy(np.stack([f[k] for f in items]))
            else:
                batch[k] = torch.tensor([f[k] for f in items])

        return cls(**batch)



from typing import Dict, List, Union

import numpy as np
import torch

from .abstract_target_transform import AbstractTargetTransform


class MultiLabelEncoder(AbstractTargetTransform):
    def __init__(self, threshold: float, labels: List[str]) -> None:
        """Multi-label encoder for multi-label classification targets.

        Args:
            threshold: Class-wise prediction threshold.
            labels: List of target labels.

        Raises:
            ValueError: If threshold is not between 0 and 1.
        """
        if not 0 <= threshold <= 1:
            raise ValueError(
                f"Threshold '{threshold}' must be between 0 and 1."
            )
        self.threshold = threshold
        self.labels = labels

    def __len__(self) -> int:
        """Get the number of unique target labels.

        Returns:
            Number of unique target labels.
        """
        return len(self.labels)

    def encode(self, x: Union[List[int], List[str]]) -> torch.Tensor:
        """Encode a list of target labels by creating a binary tensor where
        each element is 1 if the label is in the list of target labels and 0
        otherwise.

        If the input is already a list of integers, it is not encoded.

        Args:
            x: List of target labels or list of integers.

        Returns:
            Binary tensor of encoded target labels.
        """
        if len(x) == 0:
            return torch.zeros(len(self.labels))
        if isinstance(x, (torch.Tensor, np.ndarray)) or all(
            isinstance(i, (int, float)) for i in x
        ):
            return torch.Tensor(x)
        return torch.Tensor([label in x for label in self.labels])

    def decode(self, x: List[int]) -> List[str]:
        """Decode a binary tensor of encoded target labels to a list of target
        labels.

        Args:
            x: Binary tensor of encoded target labels.

        Returns:
            List of target labels.
        """
        return [label for i, label in enumerate(self.labels) if x[i]]

    def probabilities_training(self, x: torch.Tensor) -> torch.Tensor:
        """Get the encoded probabilities from a batch of model outputs
        during training by returning the raw model outputs.

        Args:
            x: Batch of model outputs.

        Returns:
            Encoded probabilities.
        """
        return x

    def probabilities_inference(self, x: torch.Tensor) -> torch.Tensor:
        """Get the encoded probabilities from a batch of model outputs by
        applying the sigmoid function.

        Args:
            x: Batch of model outputs.

        Returns:
            Encoded probabilities.
        """
        return torch.sigmoid(x)

    def predict_inference(
        self, x: torch.Tensor
    ) -> Union[List[List[int]], List[int]]:
        """Get the encoded predictions from a batch of model output
        probabilities by thresholding the probabilities.

        Args:
            x: Batch of model output probabilities.

        Returns:
            Binary tensor of encoded predictions.
        """
        return (x > self.threshold).int().squeeze().tolist()

    def majority_vote(self, x: List[List[str]]) -> List[str]:
        """Get the majority vote from a list of lists of decoded target labels
        for each label. If a label is predicted by at least half of the
        predictions, it is included in the majority vote.

        Args:
            x: List of lists of decoded target labels.

        Returns:
            List of target labels in the majority vote.
        """
        x = np.array([self.encode(i).tolist() for i in x])
        x = np.mean(x, axis=0).round().astype(int).tolist()
        return self.decode(x)

    def probabilities_to_dict(self, x: torch.Tensor) -> Dict[str, float]:
        """Convert a tensor of probabilities to a dictionary of labels and
        their probabilities.

        Args:
            x: Tensor of probabilities.

        Returns:
            Dictionary of labels and their probabilities.
        """
        return {label: prob.item() for label, prob in zip(self.labels, x)}



from collections import defaultdict
from copy import deepcopy
import os
import shutil
from typing import List, Optional

from omegaconf import DictConfig, OmegaConf
import pandas as pd

import autrainer
from autrainer.core.constants import NamingConstants
from autrainer.core.plotting import PlotMetrics
from autrainer.core.utils import Timer
from autrainer.loggers import AbstractLogger
from autrainer.metrics import AbstractMetric

from .postprocessing_utils import (
    get_plotting_params,
    get_run_names,
    get_training_type,
    load_yaml,
    save_yaml,
)
from .summarize import SummarizeGrid


class AggregateGrid:
    def __init__(
        self,
        results_dir: str,
        experiment_id: str,
        aggregate_list: List[str],
        aggregate_prefix: str = "agg",
        training_dir: str = "training",
        max_runs_plot: Optional[int] = None,
        aggregate_name: Optional[str] = None,
        aggregated_dict: Optional[dict] = None,
        plot_params: Optional[dict] = None,
    ) -> None:
        """Aggregate the results of a grid search over one or more parameters.

        If loggers have been used for the grid search, the aggregated results
        will be logged to the same loggers.

        Args:
            results_dir: The directory where the results are stored.
            experiment_id: The ID of the grid search experiment.
            aggregate_list: The list of parameters to aggregate over.
            aggregate_prefix: The prefix for the aggregated experiment ID.
                Defaults to "agg".
            training_dir: The directory of the training results of the
                experiment. Defaults to "training".
            max_runs_plot: The maximum number of best runs to plot. If None,
                all runs will be plotted. Defaults to None.
            aggregate_name: The name of the aggregated experiment. If None,
                it will be generated from the aggregate_list. Defaults to None.
            aggregated_dict: A dictionary mapping the aggregated experiment
                names to the runs to aggregate. If None, the runs will be
                aggregated based on the aggregate_list. Defaults to None.
            plot_params: Additional parameters for plotting. Defaults to None.
        """
        self.aggregate_list = aggregate_list
        self.results_dir = results_dir
        self.experiment_id = experiment_id
        self.aggregate_name = aggregate_name or "_".join(
            (aggregate_prefix, *self.aggregate_list)
        )
        self.aggregated_dict = aggregated_dict
        self.plot_params = plot_params
        self.output_directory = os.path.join(
            self.results_dir, self.experiment_id, self.aggregate_name
        )
        self.training_directory = os.path.join(
            self.results_dir, self.experiment_id, training_dir
        )
        self.max_runs_plot = max_runs_plot
        if os.path.exists(self.output_directory):
            shutil.rmtree(self.output_directory)
        os.makedirs(self.output_directory, exist_ok=True)
        self.run_names = get_run_names(self.training_directory)
        self.training_type = get_training_type(
            self.training_directory, self.run_names
        )
        self.run_names.sort()
        if self.plot_params is None:
            self.plot_params = get_plotting_params(
                self.training_directory, self.run_names[0]
            )

    def aggregate(self) -> None:
        """Aggregate the runs based on the specified parameters."""
        if self.aggregated_dict is not None:
            aggregated_runs = self.aggregated_dict
        else:
            aggregated_runs = self._aggregate_run_names(self.aggregate_list)
        for agg_name, run_list in aggregated_runs.items():
            self._aggregate_best(agg_name, run_list)
            self._aggregate_test(agg_name, run_list)
            self._aggregate_config(agg_name, run_list)
            self._aggregate_timer(agg_name, run_list)
            self._aggregate_metrics(agg_name, run_list)
            save_yaml(
                os.path.join(self.output_directory, agg_name, "runs.yaml"),
                run_list,
            )

    def summarize(self) -> None:
        """Summarize the aggregated results."""
        sg = SummarizeGrid(
            results_dir=self.results_dir,
            experiment_id=self.experiment_id,
            training_dir=self.aggregate_name,
            summary_dir=self.aggregate_name,
            clear_old_outputs=False,
            training_type=self.training_type,
            max_runs_plot=self.max_runs_plot,
            plot_params=self.plot_params,
        )
        sg.summarize()
        sg.plot_metrics()

    def _check_if_valid_aggregation(self, over: list) -> None:
        for o in over:
            if o not in NamingConstants().VALID_AGGREGATIONS:
                raise ValueError(f"Invalid aggregation: {o}")

    def _aggregate_run_names(self, over: list) -> dict:
        self._check_if_valid_aggregation(over)
        param_dict = {
            p: i for i, p in enumerate(NamingConstants().NAMING_CONVENTION)
        }
        over_idxs = [param_dict[p] for p in over]
        aggregated = defaultdict(list)
        for run_name in self.run_names:
            params = run_name.split("_")
            for idx in over_idxs:
                params[idx] = "#"
            agg_key = "_".join(params)
            aggregated[agg_key].append(run_name)
        return aggregated

    def _aggregate_best(self, agg_name: str, run_list: list):
        os.makedirs(
            os.path.join(self.output_directory, agg_name, "_best"),
            exist_ok=True,
        )
        metrics = self._aggregate_yaml(run_list, "_best/dev.yaml", "dev")
        save_yaml(
            os.path.join(self.output_directory, agg_name, "_best", "dev.yaml"),
            metrics,
        )

    def _aggregate_test(self, agg_name: str, run_list: list):
        path = os.path.join(self.output_directory, agg_name, "_test")
        os.makedirs(path, exist_ok=True)
        metrics = self._aggregate_yaml(
            run_list, "_test/test_holistic.yaml", "test"
        )
        save_yaml(os.path.join(path, "test_holistic.yaml"), metrics)

    def _aggregate_yaml(self, run_list: list, path: str, yaml_type: str):
        assert yaml_type in ["dev", "test"]
        loss_type = "dev_loss" if yaml_type == "dev" else "loss"
        dfs = []
        for run in run_list:
            metrics = load_yaml(
                os.path.join(self.training_directory, run, path)
            )
            dfs.append(pd.DataFrame(metrics))
        df = pd.concat(dfs, keys=run_list, names=["run", "type"])
        df_mean = df.groupby(level="type").mean().reset_index()
        df_std = df.groupby(level="type").std().fillna(0).reset_index()
        df_std["type"] = df_std["type"].apply(lambda x: f"{x}.std")
        df = pd.concat([df_mean, df_std]).set_index(["type"])
        metrics = df.to_dict()
        metrics[loss_type] = {
            k: v for k, v in metrics[loss_type].items() if "all" == k
        }
        if yaml_type == "dev":
            metrics["iteration"] = {
                k: v for k, v in metrics["iteration"].items() if "all" == k
            }
        return metrics

    def _aggregate_config(self, agg_name: str, run_list: list):
        path = os.path.join(self.output_directory, agg_name, ".hydra")
        os.makedirs(path, exist_ok=True)
        runs = [
            OmegaConf.load(
                os.path.join(
                    self.training_directory, r, ".hydra", "config.yaml"
                )
            )
            for r in run_list
        ]
        config = runs.pop(0)
        if self.aggregated_dict is None:
            for key in self.aggregate_list:
                config[key] = self._replace_differing_values(config, runs, key)
        save_yaml(
            os.path.join(path, "config.yaml"), OmegaConf.to_container(config)
        )

    def _replace_differing_values(
        self,
        config: DictConfig,
        runs: List[DictConfig],
        key: str,
    ) -> DictConfig:
        if not runs:
            return config[key]

        if not isinstance(config[key], DictConfig):
            return "#"

        for k in config[key].keys():
            if not isinstance(config[key][k], DictConfig):
                values = [r[key][k] for r in runs + [config]]
                if len(set(values)) > 1:
                    config[key][k] = "#"
            else:
                config[key][k] = self._replace_differing_values(
                    config[key], [r[key] for r in runs], k
                )

        return config[key]

    def _aggregate_timer(self, agg_name: str, run_list: list):
        mean_timer = {
            "train": {"mean_seconds": 0, "total_seconds": 0},
            "dev": {"mean_seconds": 0, "total_seconds": 0},
            "test": {"mean_seconds": 0, "total_seconds": 0},
        }
        for run in run_list:
            timers = load_yaml(
                os.path.join(self.training_directory, run, "timer.yaml")
            )
            for k, v in timers.items():
                mean_timer[k]["mean_seconds"] += v["mean_seconds"]
                mean_timer[k]["total_seconds"] += v["total_seconds"]
        for k, v in mean_timer.items():
            v["mean_seconds"] /= len(run_list)
            v["mean"] = Timer.pretty_time(v["mean_seconds"])
            v["total_seconds"] /= len(run_list)
            v["total"] = Timer.pretty_time(v["total_seconds"])
        save_yaml(
            os.path.join(self.output_directory, agg_name, "timer.yaml"),
            mean_timer,
        )

    def _aggregate_metrics(self, agg_name: str, run_list: list):
        dfs = []
        for run in run_list:
            df = pd.read_csv(
                os.path.join(self.training_directory, run, "metrics.csv"),
                index_col="iteration",
            )
            dfs.append(df)
        cfg = OmegaConf.load(
            os.path.join(
                self.output_directory, agg_name, ".hydra", "config.yaml"
            )
        )
        if isinstance(cfg.dataset.metrics, str):
            raise ValueError(
                "Unable to aggregate over datasets with different metrics."
            )
        metric_fns = [
            autrainer.instantiate_shorthand(m, instance_of=AbstractMetric)
            for m in cfg.dataset.metrics
        ]
        tracking_metric = autrainer.instantiate_shorthand(
            config=cfg.dataset.tracking_metric,
            instance_of=AbstractMetric,
        )

        df = pd.concat(dfs, keys=run_list, names=["run", "iteration"])
        mean_df = df.groupby(level="iteration").mean()
        std_df = df.groupby(level="iteration").std().fillna(0)
        df = mean_df.join(std_df, rsuffix=".std")
        df["iteration"] = df.index
        df.to_csv(
            os.path.join(self.output_directory, agg_name, "metrics.csv"),
            index=False,
        )
        df.drop(columns=["iteration"], inplace=True)

        plotter = PlotMetrics(
            output_directory=os.path.join(self.output_directory, agg_name),
            training_type=self.training_type,
            metric_fns=metric_fns,
            **self.plot_params,
        )
        plotter.plot_run(df)

        loggers = []
        for logger in cfg.get("loggers", []):
            loggers.append(
                autrainer.instantiate_shorthand(
                    logger,
                    instance_of=AbstractLogger,
                    exp_name=self.experiment_id + "." + self.aggregate_name,
                    run_name=agg_name,
                    metrics=metric_fns,
                    tracking_metric=tracking_metric,
                )
            )
        timers = load_yaml(
            os.path.join(self.output_directory, agg_name, "timer.yaml")
        )
        test_metrics = load_yaml(
            os.path.join(
                self.output_directory, agg_name, "_test", "test_holistic.yaml"
            )
        )
        for logger in loggers:
            logger.setup()
            logger.log_params(deepcopy(cfg))
            logger.log_timers(
                {
                    "time.train.mean": timers["train"]["mean"],
                    "time.dev.mean": timers["dev"]["mean"],
                    "time.test.mean": timers["test"]["mean"],
                }
            )
            for iteration in df.index:
                metrics = df.loc[iteration].to_dict()
                metrics = {
                    k: v for k, v in metrics.items() if not k.endswith(".std")
                }
                logger.log_metrics(metrics, iteration)
            logger.log_metrics(
                {"test_" + k: v["all"] for k, v in test_metrics.items()}
            )
            logger.log_artifact(
                os.path.join(agg_name, ".hydra", "config.yaml"),
                self.output_directory,
            )
            logger.log_artifact(
                os.path.join(agg_name, "metrics.csv"), self.output_directory
            )
            logger.end_run()



from typing import Callable

import audmetric
import numpy as np

from .abstract_metric import BaseAscendingMetric, BaseDescendingMetric


def _mean_metric(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    metric_fn: Callable,
    **kwargs,
) -> float:
    """Calculate the mean of a metric for each target if multi-target arrays
    are given. If single-target arrays are given, the metric is calculated
    directly.

    Args:
        y_true: Ground truth values.
        y_pred: Predicted values.
        metric_fn: Metric function.

    Returns:
        Mean of the metric for each target.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    if y_true.ndim == 1:
        return metric_fn(y_true, y_pred, **kwargs)
    m = np.empty(y_true.shape[1])
    for i in range(y_true.shape[1]):
        m[i] = metric_fn(y_true[:, i], y_pred[:, i], **kwargs)
    return np.mean(m)


class MSE(BaseDescendingMetric):
    def __init__(self):
        """Mean squared error metric using `audmetric.mean_squared_error`
        for (multi-target) regression. The metric is calculated for each
        target separately and the mean is returned.
        """
        super().__init__(
            name="mse",
            fn=_mean_metric,
            metric_fn=audmetric.mean_squared_error,
        )


class MAE(BaseDescendingMetric):
    def __init__(self):
        """Mean absolute error metric using `audmetric.mean_absolute_error`
        for (multi-target) regression. The metric is calculated for each
        target separately and the mean is returned.
        """
        super().__init__(
            name="mae",
            fn=_mean_metric,
            metric_fn=audmetric.mean_absolute_error,
        )


class PCC(BaseAscendingMetric):
    def __init__(self):
        """Pearson correlation coefficient metric using
        `audmetric.pearson_cc` for (multi-target) regression. The metric is
        calculated for each target separately and the mean is returned.
        """
        super().__init__(
            name="pcc",
            fn=_mean_metric,
            metric_fn=audmetric.pearson_cc,
        )


class CCC(BaseAscendingMetric):
    def __init__(self):
        """Concordance correlation coefficient metric using
        `audmetric.concordance_cc` for (multi-target) regression. The metric is
        calculated for each target separately and the mean is returned.
        """
        super().__init__(
            name="ccc",
            fn=_mean_metric,
            metric_fn=audmetric.concordance_cc,
        )



import torch
import torchvision.transforms.functional as F


def _get_flat_grid_locations(height, width, device):
    y_range = torch.linspace(0, height - 1, height, device=device)
    x_range = torch.linspace(0, width - 1, width, device=device)
    y_grid, x_grid = torch.meshgrid(y_range, x_range, indexing="ij")
    return torch.stack((y_grid, x_grid), -1).reshape([height * width, 2])


def _phi(r, order):
    EPSILON = torch.tensor(1e-10, device=r.device)
    if order == 1:
        r = torch.max(r, EPSILON)
        r = torch.sqrt(r)
        return r
    elif order == 2:
        return 0.5 * r * torch.log(torch.max(r, EPSILON))
    elif order == 4:
        return 0.5 * torch.square(r) * torch.log(torch.max(r, EPSILON))
    elif order % 2 == 0:
        r = torch.max(r, EPSILON)
        return 0.5 * torch.pow(r, 0.5 * order) * torch.log(r)
    else:
        r = torch.max(r, EPSILON)
        return torch.pow(r, 0.5 * order)


def _cross_squared_distance_matrix(x, y):
    x_norm_squared = torch.sum(torch.mul(x, x))
    y_norm_squared = torch.sum(torch.mul(y, y))

    x_y_transpose = torch.matmul(x.squeeze(0), y.squeeze(0).transpose(0, 1))

    squared_dists = x_norm_squared - 2 * x_y_transpose + y_norm_squared

    return squared_dists.float()


def _solve_interpolation(pts, vals, order, eps: float = 1e-7) -> tuple:
    device = pts.device
    channels, n, d = F.get_dimensions(pts)
    k = vals.shape[-1]

    c = pts[0]
    f = vals[0].float()

    matrix_a = _phi(_cross_squared_distance_matrix(c, c), order)  # [n, n]
    ones = torch.ones(n, dtype=pts.dtype, device=device).view([n, 1])  # [n ,1]
    matrix_b = torch.cat((c, ones), -1).float()  # [n , d + 1]

    # [n + d + 1, n]
    left_block = torch.cat((matrix_a, torch.transpose(matrix_b, 1, 0)), 0)

    num_b_cols = matrix_b.shape[-1]

    lhs_zeros = torch.randn((num_b_cols, num_b_cols), device=device) * eps
    right_block = torch.cat((matrix_b, lhs_zeros), 0)  # [n + d + 1, d + 1]
    lhs = torch.cat((left_block, right_block), 1)  # [n + d + 1, n + d + 1]

    rhs_zeros = torch.zeros((d + 1, k), dtype=pts.dtype, device=device).float()
    rhs = torch.cat((f, rhs_zeros), 0)  # [n + d + 1, k]

    try:
        X = torch.linalg.solve(lhs, rhs)
        w = X[:n, :]
        v = X[n:, :]
        return w, v
    except torch._C._LinAlgError:
        return None, None


def _apply_interpolation(query_pts, pts, w, v, order):
    query_pts = query_pts.unsqueeze(0)
    pairwise_dists = _cross_squared_distance_matrix(
        query_pts.float(), pts.float()
    )
    phi_pairwise_dists = _phi(pairwise_dists, order)

    rbf_term = torch.matmul(phi_pairwise_dists, w)

    ones = torch.ones_like(query_pts[..., :1])
    query_points_pad = torch.cat((query_pts, ones), 2).float()
    linear_term = torch.matmul(query_points_pad, v)

    return rbf_term + linear_term


def _interpolate_spline(
    pts,
    vals,
    query_pts,
    order,
    regularization_weight=0.0,
) -> torch.Tensor:
    # First, fit the spline to the observed data.
    w, v = _solve_interpolation(pts, vals, order, regularization_weight)

    if w is None and v is None:
        return None

    # Then, evaluate the spline at the query locations.
    query_values = _apply_interpolation(query_pts, pts, w, v, order)

    return query_values


def _create_dense_flows(flattened_flows, height, width):
    # possibly .view
    return torch.reshape(flattened_flows, [height, width, 2])


def _interpolate_bilinear(
    grid, query_points, name="interpolate_bilinear", indexing="ij"
):
    if indexing != "ij" and indexing != "xy":
        raise ValueError("Indexing mode must be 'ij' or 'xy'")

    shape = grid.shape
    if len(shape) != 3:
        msg = "Grid must be 3 dimensional. Received size: "
        raise ValueError(msg + str(grid.shape))

    channels, height, width = F.get_dimensions(grid)

    shape = [height, width, channels]
    query_type = query_points.dtype
    grid_type = grid.dtype
    grid_device = grid.device

    num_queries = query_points.shape[0]

    alphas = []
    floors = []
    ceils = []
    index_order = [0, 1] if indexing == "ij" else [1, 0]
    unstacked_query_points = query_points.unbind(1)

    for dim in index_order:
        queries = unstacked_query_points[dim]

        size_in_indexing_dimension = shape[dim]

        max_floor = torch.tensor(
            size_in_indexing_dimension - 2,
            dtype=query_type,
            device=grid_device,
        )
        min_floor = torch.tensor(0, dtype=query_type, device=grid_device)
        maxx = torch.max(min_floor, torch.floor(queries))
        floor = torch.min(maxx, max_floor)
        int_floor = floor.long()
        floors.append(int_floor)
        ceil = int_floor + 1
        ceils.append(ceil)

        alpha = (queries - floor).clone().detach().type(grid_type)
        min_alpha = torch.tensor(0, dtype=grid_type, device=grid_device)
        max_alpha = torch.tensor(1, dtype=grid_type, device=grid_device)
        alpha = torch.min(torch.max(min_alpha, alpha), max_alpha)

        alpha = torch.unsqueeze(alpha, 1)
        alphas.append(alpha)

    # work with only one channel
    flattened_grid = torch.reshape(grid[0], [height * width, 1])

    def gather(y_coords, x_coords, name):
        linear_coordinates = y_coords * width + x_coords
        gathered_values = torch.gather(
            flattened_grid.t(), 1, linear_coordinates.unsqueeze(0)
        )
        return torch.reshape(gathered_values, [num_queries, 1])

    # grab the pixel values in the 4 corners around each query point
    top_left = gather(floors[0], floors[1], "top_left")
    top_right = gather(floors[0], ceils[1], "top_right")
    bottom_left = gather(ceils[0], floors[1], "bottom_left")
    bottom_right = gather(ceils[0], ceils[1], "bottom_right")

    interp_top = alphas[1] * (top_right - top_left) + top_left
    interp_bottom = alphas[1] * (bottom_right - bottom_left) + bottom_left
    interp = alphas[0] * (interp_bottom - interp_top) + interp_top

    interp = interp.repeat(channels, 1, 1)

    return interp


def _dense_image_warp(tensor, flow):
    channels, height, width = F.get_dimensions(tensor)
    device = tensor.device

    grid_x, grid_y = torch.meshgrid(
        torch.arange(width, device=device),
        torch.arange(height, device=device),
        indexing="ij",
    )

    stacked_grid = torch.stack((grid_y, grid_x), dim=2).float()
    b_grid = stacked_grid.unsqueeze(-1).permute(3, 1, 0, 2)
    query_points_on_grid = b_grid - flow

    query_points_flattened = torch.reshape(
        query_points_on_grid, [height * width, 2]
    )

    interpolated = _interpolate_bilinear(tensor, query_points_flattened)
    interpolated = torch.reshape(interpolated, [channels, height, width])

    return interpolated


def _sparse_image_warp(
    tensor,
    src_ctr_pt_locations,
    dest_ctr_pt_locations,
    order=2,
    regularization_weight=0.0,
):
    device = tensor.device
    control_point_flows = dest_ctr_pt_locations - src_ctr_pt_locations

    _, height, width = F.get_dimensions(tensor)
    flattened_grid_locations = _get_flat_grid_locations(height, width, device)

    flattened_flows = _interpolate_spline(
        dest_ctr_pt_locations,
        control_point_flows,
        flattened_grid_locations,
        order,
        regularization_weight,
    )

    if flattened_flows is None:
        return tensor, None

    dense_flows = _create_dense_flows(flattened_flows, height, width)

    warped_image = _dense_image_warp(tensor, dense_flows)
    return warped_image, dense_flows



from abc import ABC, abstractmethod
import argparse
import sys
from typing import Any, Dict, List, Optional, Type, TypeVar

from autrainer.core.constants import NamingConstants

from .command_line_error import CommandLineError


T = TypeVar("T")


class AbstractScript(ABC):
    """Abstract script class.

    Args:
        command (str): Command name.
        description (str): Command description.
        extended_description (str, optional): Extended description.
        epilog (str, optional): Command epilog. Defaults to "".
        dataclass (Type[T], optional): Dataclass type. Defaults to None.
        unknown_args (bool, optional): Allow to pass unknown arguments.
        Defaults to False.
    """

    def __init__(
        self,
        command: str,
        description: str,
        extended_description: str = "",
        epilog: str = "",
        dataclass: Optional[Type[T]] = None,
        unknown_args: bool = False,
    ):
        self.command = command
        self.description = description
        self.extended_description = extended_description
        self.epilog = epilog
        self.dataclass = dataclass or dict
        self.unknown_args = unknown_args

    def init_parser(self, parser: argparse.ArgumentParser) -> None:
        """Initialize the parser.

        Args:
            parser (argparse.ArgumentParser): The parser.
        """
        self.parser = parser

    def add_arguments(self) -> None:
        """Add arguments to the parser."""

    def run(self, args: Dict[str, Any], unknown: List[str]) -> None:
        """Run the script.

        Args:
            args (Dict[str, Any]): Parsed arguments.
            unknown (List[str]): List of unknown arguments.
        """
        assert getattr(self, "parser", None) is not None
        if not self.unknown_args:
            self._assert_no_unknown_args(unknown)
        else:
            sys.argv = sys.argv[:1] + unknown
        if self.dataclass is not dict:
            args = self._create_dataclass(args, self.dataclass)
        self.main(args)

    @abstractmethod
    def main(self, args: T) -> None:
        """Main function to run the script.

        Args:
            args (T): Parsed arguments as a dataclass instance or dict if no
            dataclass is provided.
        """

    def _create_dataclass(self, args: Dict[str, Any], dataclass: Type[T]) -> T:
        """Create a dataclass instance from parsed arguments.

        Args:
            args (Dict[str, Any]): Parsed arguments.
            dataclass (T): Dataclass type.

        Returns:
            T: Dataclass instance.
        """
        return dataclass(**args)

    def _assert_config_directory(self, directory: str) -> None:
        """Assert that the configuration directory is valid.

        Args:
            directory (str): Configuration directory.

        Raises:
            CommandLineError: If the configuration directory is invalid.
        """
        if directory not in NamingConstants().CONFIG_DIRS:
            raise CommandLineError(
                parser=self.parser,
                message=f"Invalid configuration directory '{directory}'.",
            )

    def _assert_no_unknown_args(self, unknown: List[str]) -> None:
        """Assert that there are no unknown arguments.

        Args:
            unknown (List[str]): List of unknown arguments.

        Raises:
            CommandLineError: If there are unknown arguments.
        """
        if unknown:
            raise CommandLineError(
                parser=self.parser,
                message=f"Unknown arguments: {unknown}.",
            )


class MockParser:
    def add_argument(self, *args, **kwargs) -> None:
        pass

    def print_help(self, *args, **kwargs) -> None:
        pass



import os
import sys
from typing import Optional

import hydra
from hydra.core.config_search_path import ConfigSearchPath
from hydra.core.plugins import Plugins
from hydra.plugins.search_path_plugin import SearchPathPlugin

import autrainer

from .config_dir_snapshot import ConfigDirSnapshot


class AutrainerPathPlugin(SearchPathPlugin):
    def manipulate_search_path(self, search_path: ConfigSearchPath) -> None:
        lib_path = os.path.join(
            os.path.dirname(autrainer.__path__[0]),
            "autrainer-configurations",
        )
        snapshot = ConfigDirSnapshot().create_snapshot_dir()
        if snapshot is not None:
            search_path.append(
                provider="autrainer-snapshot",
                path=f"file://{snapshot}",
            )
        search_path.append(
            provider="autrainer-configs",
            path=f"file://{lib_path}",
        )


def add_current_directory_to_path() -> None:
    current_dir = os.getcwd()
    if current_dir not in sys.path:
        sys.path.insert(0, current_dir)


def main(
    config_name: str,
    config_path: Optional[str] = None,
    version_base: Optional[str] = None,
):
    """Hydra main decorator with additional `autrainer` configs.

    The `conf` directory in the current working directory is always added to
    the search path if it exists.
    The current working directory is also added to the Python path.

    Args:
        config_name: The name of the config (usually the file name without the
            .yaml extension).
        config_path: The config path, a directory where Hydra will search for
            config files. If config_path is None no directory is added to the
            search path. Defaults to None.
        version_base: Hydra version base. Defaults to None.
    """
    if not any("jupyter" in arg or "ipykernel" in arg for arg in sys.argv):
        import matplotlib

        matplotlib.use("Agg")  # TkAgg is not thread-safe

    Plugins.instance().register(AutrainerPathPlugin)
    add_current_directory_to_path()
    return hydra.main(
        version_base=version_base,
        config_path=config_path,
        config_name=config_name,
    )



from typing import Any, Dict, Tuple, Union


def convert_shorthand(aug: Union[str, Dict[str, Any]]) -> Dict[str, Any]:
    """Convert shorthand syntax to a dictionary.

    Args:
        aug: The augmentation to convert.

    Returns:
        The converted augmentation.
    """
    if isinstance(aug, str):
        return {aug: {}}
    return aug


def assign_seeds(
    aug: Dict[str, Any],
    current_seed: int,
    increment: bool,
) -> Tuple[Dict[str, Any], int]:
    """Recursively assign optionally incrementing seeds to augmentation
    configurations.

    Args:
        aug: Potentially nested augmentation configuration.
        current_seed: The current seed to assign.
        increment: Whether to increment the seed after assignment. If True,
            the seed of each augmentation will be incremented by 1.
            If False, the same seed will be used for all augmentations.

    Returns:
        The updated augmentation configuration and the next seed to use.
    """
    aug_name = next(iter(aug))
    config = aug[aug_name]

    if config.get("generator_seed") is None:
        config["generator_seed"] = current_seed
        if increment:
            current_seed += 1

    # recursively assign unique seeds to each augmentation in choice or
    # sequential to avoid shared seeds, which can lead to identical
    # augmentations within a batch
    if aug_name == "autrainer.augmentations.Choice" and "choices" in config:
        new_choices = []
        for choice in config["choices"]:
            choice = convert_shorthand(choice)
            updated_choice, current_seed = assign_seeds(
                choice, current_seed, increment
            )
            new_choices.append(updated_choice)
        config["choices"] = new_choices

    elif (
        aug_name == "autrainer.augmentations.Sequential"
        and "sequence" in config
    ):
        new_sequence = []
        for seq_aug in config["sequence"]:
            seq_aug = convert_shorthand(seq_aug)
            updated_aug, current_seed = assign_seeds(
                seq_aug, current_seed, increment
            )
            new_sequence.append(updated_aug)
        config["sequence"] = new_sequence

    return {aug_name: config}, current_seed



import logging
import os
import sys
from typing import Dict, List, Optional, Tuple, Union
import warnings

import audobject
import numpy as np
import pandas as pd
import torch
from torchinfo import summary
import yaml

from autrainer.metrics import AbstractMetric


class Bookkeeping:
    def __init__(
        self,
        output_directory: str,
        file_handler_path: Optional[str] = None,
    ) -> None:
        """Bookkeeping to handle general disk operations and interactions.

        Args:
            output_directory: Output directory to save files to.
            file_handler_path: Path to save the log file to. Defaults to None.
        """
        self.output_directory = output_directory
        os.makedirs(self.output_directory, exist_ok=True)
        self.original_stdout = sys.stdout
        # ? Setup Custom Logging
        self.logger = logging.getLogger()
        if not self.logger.hasHandlers() or file_handler_path is not None:
            self._setup_logger(file_handler_path)

        for handler in self.logger.handlers:
            if isinstance(handler, logging.StreamHandler):
                handler.setFormatter(
                    logging.Formatter(
                        "[%(asctime)s][%(levelname)s]\n%(message)s\n"
                    )
                )

    def _setup_logger(self, fp: Optional[str] = None) -> None:
        self.logger.setLevel(logging.INFO)
        if not self.logger.hasHandlers():
            self.logger.addHandler(logging.StreamHandler())
        if fp is not None:
            self.logger.addHandler(logging.FileHandler(fp))
        else:
            self.logger.addHandler(
                logging.FileHandler(
                    os.path.join(self.output_directory, "bookkeeping.log")
                )
            )

    def log(self, message: str, level: int = logging.INFO) -> None:
        """Log a message.

        Args:
            message: Message to log.
            level: Logging level. Defaults to logging.INFO.
        """
        self.logger.log(level, message)

    def log_to_file(self, message: str, level: int = logging.INFO) -> None:
        """Log a message to the file handler.

        Args:
            message: Message to log.
            level: Logging level. Defaults to logging.INFO.
        """
        for handler in self.logger.handlers:
            if isinstance(handler, logging.FileHandler):
                handler.emit(
                    logging.LogRecord(
                        self.logger.name,
                        level,
                        None,
                        None,
                        message,
                        None,
                        None,
                    )
                )

    def create_folder(self, folder_name: str, path: str = "") -> None:
        """Create a new folder in the output directory.

        Args:
            folder_name: Name of the folder to create.
            path: Subdirectory to create the folder in. Defaults to "".
        """
        os.makedirs(
            os.path.join(self.output_directory, path, folder_name),
            exist_ok=True,
        )

    def save_model_summary(
        self,
        model: torch.nn.Module,
        shape: Tuple[int, ...],
        device: torch.device,
        filename: str,
    ) -> None:
        """Save a model summary to a file.

        Args:
            model: Model to summarize.
            shape: Shape of the input to the model.
            filename: Name of the file to save the summary to.
        """
        s = summary(
            model=model,
            device=device,
            input_size=(shape),
            col_names=["input_size", "output_size", "num_params", "trainable"],
            row_settings=["var_names"],
            verbose=0,
        )

        model_summary = {
            "total_mult_adds": s.total_mult_adds,
            "total_output_bytes": s.total_output_bytes,
            "total_params": s.total_params,
            "trainable_params": s.trainable_params,
            "total_param_bytes": s.total_param_bytes,
        }

        out_path = os.path.join(self.output_directory, filename)
        with open(out_path, "w", encoding="utf-8") as f:
            f.write(str(s))

        with open(out_path.replace(".txt", ".yaml"), "w") as f:
            yaml.dump(model_summary, f)

    def save_state(
        self,
        obj: Optional[
            Union[
                torch.nn.Module,
                torch.optim.Optimizer,
                torch.optim.lr_scheduler.LRScheduler,
            ]
        ],
        filename: str,
        path: str = "",
    ) -> None:
        """Save the state of an object.

        Args:
            obj: Object to save the state of. If None, do nothing.
            filename: Name of the file to save the state to.
            path: Subdirectory to save the state to. Defaults to "".

        Raises:
            TypeError: If the object type is not supported.
        """
        if obj is None:
            return
        p = os.path.join(self.output_directory, path, filename)
        _i = (
            torch.nn.Module,
            torch.optim.Optimizer,
            torch.optim.lr_scheduler.LRScheduler,
        )
        if not isinstance(obj, _i):
            raise TypeError(
                f"save_state of type {type(obj)} is not supported."
            )

        os.makedirs(os.path.join(self.output_directory, path), exist_ok=True)
        torch.save(obj.state_dict(), p)

    def load_state(
        self,
        obj: Union[
            torch.nn.Module,
            torch.optim.Optimizer,
            torch.optim.lr_scheduler.LRScheduler,
        ],
        filename: str,
        path: str = "",
    ) -> None:
        """Load the state of an object.

        Args:
            obj: Object to load the state into.
            filename: Name of the file to load the state from.
            path: Subdirectory to load the state from. Defaults to "".

        Raises:
            TypeError: If the object type is not supported.
            FileNotFoundError: If the file is not found.
        """
        p = os.path.join(self.output_directory, path, filename)
        if not os.path.isfile(p):
            raise FileNotFoundError(f"File {p} not found.")
        _i = (
            torch.nn.Module,
            torch.optim.Optimizer,
            torch.optim.lr_scheduler.LRScheduler,
        )
        if not isinstance(obj, _i):
            raise TypeError(
                f"load_state of type {type(obj)} is not supported."
            )

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            state_dict = torch.load(p, map_location="cpu", weights_only=True)
        obj.load_state_dict(state_dict)

    def save_audobject(
        self,
        obj: audobject.Object,
        filename: str,
        path: str = "",
    ) -> None:
        """Save an audobject.Object to disk.

        Args:
            obj: Object to save.
            filename: Name of the file to save the object to.
            path: Subdirectory to save the object to. Defaults to "".

        Raises:
            TypeError: If the object type is not supported.
        """
        if not isinstance(obj, audobject.Object):
            raise TypeError(
                f"save_audobject of type {type(obj)} is not supported."
            )
        os.makedirs(os.path.join(self.output_directory, path), exist_ok=True)
        obj.to_yaml(os.path.join(self.output_directory, path, filename))

    def save_results_dict(
        self,
        results_dict: Dict[str, float],
        filename: str,
        path: str = "",
    ) -> None:
        """Save a results dictionary to disk.

        Args:
            results_dict: Dictionary of metric names and values to save.
            filename: Name of the file to save the results to.
            path: Subdirectory to save the results to. Defaults to "".
        """
        os.makedirs(os.path.join(self.output_directory, path), exist_ok=True)
        with open(
            os.path.join(self.output_directory, path, filename),
            "w",
            encoding="utf-8",
        ) as f:
            yaml.dump(results_dict, f)

    def save_results_df(
        self, results_df: pd.DataFrame, filename: str, path: str = ""
    ) -> None:
        """Save a results DataFrame to disk.

        Args:
            results_df: DataFrame to save.
            filename: Name of the file to save the results to.
            path: Subdirectory to save the results to. Defaults to "".
        """
        os.makedirs(os.path.join(self.output_directory, path), exist_ok=True)
        results_df.to_csv(
            os.path.join(self.output_directory, path, filename), index=False
        )

    def save_results_np(
        self, results_np: np.ndarray, filename: str, path: str = ""
    ) -> None:
        """Save a results numpy array to disk.

        Args:
            results_np: Numpy array to save.
            filename: Name of the file to save the results to.
            path: Subdirectory to save the results to. Defaults to "".
        """
        os.makedirs(os.path.join(self.output_directory, path), exist_ok=True)
        np.save(
            os.path.join(self.output_directory, path, filename), results_np
        )

    def save_best_results(
        self,
        metrics: pd.DataFrame,
        filename: str,
        metric_fns: List[AbstractMetric],
        tracking_metric_fn: AbstractMetric,
        path: str = "",
    ) -> None:
        """Save the best results to disk.

        Args:
            metrics: DataFrame of metrics to save.
            filename: Name of the file to save the best results to.
            metric_fns: List of metric functions to get the best results from.
            tracking_metric_fn: Tracking metric function to get the best
                iteration from.
            path: Subdirectory to save the best results to. Defaults to "".
        """
        best_metrics = {}
        for m in metrics:
            if "loss" in m:
                best_metrics[f"{m}_min"] = float(metrics[m].min())

        for m in metric_fns:
            best_metrics[m.name] = m.get_best(metrics[m.name])

        best_metrics["best_iteration"] = tracking_metric_fn.get_best_pos(
            metrics[tracking_metric_fn.name]
        )
        os.makedirs(os.path.join(self.output_directory, path), exist_ok=True)
        with open(
            os.path.join(self.output_directory, path, filename),
            "w",
            encoding="utf-8",
        ) as f:
            yaml.dump(best_metrics, f)



from dataclasses import dataclass
import shutil
import sys
import tempfile

from .abstract_script import AbstractScript


@dataclass
class PreprocessArgs:
    cfg_launcher: bool


class AbstractPreprocessScript(AbstractScript):
    def add_arguments(self) -> None:
        self.parser.add_argument(
            "-l",
            "--cfg-launcher",
            action="store_true",
            required=False,
            help=(
                "Use the launcher specified in the configuration instead of "
                "the Hydra basic launcher. Defaults to False."
            ),
        )

    def _override_launcher(self, args: PreprocessArgs) -> None:
        self._pre_sys_argv = sys.argv.copy()
        if not args.cfg_launcher:
            sys.argv = sys.argv + ["hydra/launcher=basic"]

        self.tempdir = tempfile.mkdtemp()
        sys.argv = sys.argv + [
            "hydra/hydra_logging=none",
            "hydra/job_logging=none",
            f"hydra.sweep.dir={self.tempdir}",
            f"hydra.sweep.subdir={self.tempdir}",
        ]

    def _clean_up(self) -> None:
        shutil.rmtree(self.tempdir)
        sys.argv = self._pre_sys_argv

    @staticmethod
    def _id_in_dict(d: dict, id: str) -> bool:
        for v in d.values():
            if v["id"] == id:
                return True
        return False



from dataclasses import dataclass

import torch

from .abstract_data_struct import AbstractDataBatch, AbstractDataItem


@dataclass
class DataItem(AbstractDataItem):
    """Data item for a single sample.

    Args:
        features: Tensor of input features.
        target: Target value for the input features.
        index: Index of the data sample.
    """


@dataclass
class DataBatch(AbstractDataBatch[DataItem]):
    """Data batch for a batch of samples.

    Args:
        features: Tensor of input features.
        target: Tensor of target values for the input features.
        index: Tensor of indices for the data samples.
    """

    def to(self, device: torch.device, **kwargs: dict) -> None:
        self.features = self.features.to(device, **kwargs)
        self.target = self.target.to(device, **kwargs)



from autrainer.core.wrappers import (
    HydraConvertEnum,
    instantiate,
    instantiate_shorthand,
    main,
)


__version__ = "0.7.1"

__all__ = ["HydraConvertEnum", "instantiate", "instantiate_shorthand", "main"]



from dataclasses import dataclass
import os
import shutil
from typing import List, Optional

import autrainer
from autrainer.core.constants import NamingConstants

from .abstract_script import AbstractScript, MockParser
from .command_line_error import CommandLineError
from .utils import catch_cli_errors


@dataclass
class CreateArgs:
    directories: List[str]
    empty: bool
    all: bool
    force: bool


class CreateScript(AbstractScript):
    def __init__(self) -> None:
        super().__init__(
            "create",
            "Create a new project with default configurations.",
            epilog="Example: autrainer create -e",
            dataclass=CreateArgs,
        )

    def add_arguments(self) -> None:
        self.parser.add_argument(
            "directories",
            type=str,
            nargs="*",
            help="Configuration directories to create. One or more of:"
            + "\n - ".join([""] + sorted(NamingConstants().CONFIG_DIRS)),
        )
        self.parser.add_argument(
            "-e",
            "--empty",
            action="store_true",
            help="Create an empty project without any configuration directory.",
        )
        self.parser.add_argument(
            "-a",
            "--all",
            action="store_true",
            help="Create a project with all configuration directories.",
        )
        self.parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            help=(
                "Force overwrite if the configuration directory "
                "already exists."
            ),
        )

    def main(self, args: CreateArgs) -> None:
        self._assert_valid_args(args)
        self._assert_config_directories(args)
        self._assert_mutually_exclusive(args)
        self._assert_directory_not_exists(args)
        self._create_directory(args)
        self._create_default_config()

    def _assert_valid_args(self, args: CreateArgs) -> None:
        if not args.directories and not args.empty and not args.all:
            raise CommandLineError(
                parser=self.parser,
                message="No configuration directories specified.",
            )

    def _assert_config_directories(self, args: CreateArgs) -> None:
        if args.empty or args.all:
            return
        for directory in args.directories:
            self._assert_config_directory(directory)

    def _assert_mutually_exclusive(self, args: CreateArgs) -> None:
        if args.empty and args.all:
            raise CommandLineError(
                parser=self.parser,
                message=(
                    "The flags -e/--empty and -a/--all "
                    "are mutually exclusive."
                ),
            )
        if args.directories and (args.empty or args.all):
            raise CommandLineError(
                parser=self.parser,
                message=(
                    "The flags -e/--empty and -a/--all "
                    "are mutually exclusive with configuration directories."
                ),
            )

    def _assert_directory_not_exists(self, args: CreateArgs) -> None:
        if args.force:
            return
        if os.path.exists("conf") and os.path.isdir("conf"):
            raise CommandLineError(
                parser=self.parser,
                message=(
                    "Directory 'conf' already exists. "
                    "Use -f/--force to overwrite it."
                ),
                code=1,
            )

    def _create_directory(self, args: CreateArgs) -> None:
        directories = []
        if args.empty:
            directories.append("conf")
        elif args.all:
            directories.extend(
                f"conf/{directory}"
                for directory in NamingConstants().CONFIG_DIRS
            )
        else:
            directories.extend(
                f"conf/{directory}" for directory in args.directories
            )

        for directory in directories:
            os.makedirs(directory, exist_ok=True)

    def _create_default_config(self) -> None:
        src = os.path.join(
            os.path.dirname(autrainer.__path__[0]),
            "autrainer-configurations",
            "config.yaml",
        )
        dst = os.path.join("conf", "config.yaml")
        shutil.copyfile(src, dst)


@catch_cli_errors
def create(
    directories: Optional[List[str]] = None,
    empty: bool = False,
    all: bool = False,
    force: bool = False,
) -> None:
    """Create a new project with default configurations.

    If called in a notebook, the function will not raise an error and print
    the error message instead.

    Args:
        directories: Configuration directories to create. One or more of:
            :const:`~autrainer.core.constants.NamingConstants.CONFIG_DIRS`.
            Defaults to None.
        empty: Create an empty project without any configuration directory.
            Defaults to False.
        all: Create a project with all configuration directories.
            Defaults to False.
        force: Force overwrite if the configuration directory already exists.
            Defaults to False.

    Raises:
        CommandLineError: If no configuration directories are specified and
            neither the empty nor all flags are set.
        CommandLineError: If the empty and all flags are set at the same time.
        CommandLineError: If the empty or all flags are set in combination with
            configuration directories.
        CommandLineError: If the configuration directory already exists and the
            force flag is not set.
    """

    script = CreateScript()
    script.parser = MockParser()
    script.main(CreateArgs(directories, empty, all, force))



from functools import cached_property
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
from omegaconf import DictConfig
import pandas as pd
import torch

from autrainer.core.constants import TrainingConstants
from autrainer.core.structs import DataItem
from autrainer.transforms import SmartCompose

from .abstract_dataset import AbstractDataset
from .utils import (
    AbstractTargetTransform,
    LabelEncoder,
    MinMaxScaler,
    MultiLabelEncoder,
    MultiTargetMinMaxScaler,
)


class ToyDatasetWrapper(torch.utils.data.Dataset):
    def __init__(
        self,
        df: pd.DataFrame,
        target_column: Union[list, str],
        feature_shape: Union[int, List[int]],
        dtype: str,
        generator: torch.Generator,
        transform: SmartCompose = None,
        target_transform: SmartCompose = None,
    ) -> None:
        self.df = df
        self.target_column = target_column
        self.feature_shape = feature_shape
        self.transform = transform
        self.target_transform = target_transform
        self.dtype = dtype
        self.generator = generator

    def __len__(self) -> int:
        return len(self.df)

    def __getitem__(self, index: int) -> Tuple[torch.Tensor, int, int]:
        target = self.df.iloc[index][self.target_column]
        if isinstance(target, pd.Series):
            target = target.values
        if self.dtype == "float32":
            data = torch.rand(
                self.feature_shape,
                dtype=torch.float32,
                generator=self.generator,
            )
        else:
            data = torch.randint(
                0,
                256,
                self.feature_shape,
                dtype=torch.uint8,
                generator=self.generator,
            )
        it = DataItem(features=data, target=target, index=index)
        if self.transform:
            it = self.transform(it)
        if self.target_transform:
            it.target = self.target_transform(it.target)
        return it


class ToyDataset(AbstractDataset):
    def __init__(
        self,
        task: str,
        size: int,
        num_targets: int,
        feature_shape: Union[int, List[int]],
        dev_split: float,
        test_split: float,
        seed: int,
        metrics: List[Union[str, DictConfig, Dict]],
        tracking_metric: Union[str, DictConfig, Dict],
        dtype: str = "float32",
        train_transform: Optional[SmartCompose] = None,
        dev_transform: Optional[SmartCompose] = None,
        test_transform: Optional[SmartCompose] = None,
    ) -> None:
        """Toy dataset for testing purposes.

        Args:
            task: Task of the dataset in ["classification", "regression",
                "ml-classification", "mt-regression"].
            size: Size of the dataset.
            num_targets: Number of targets.
            feature_shape: Shape of the features.
            dev_split: Proportion of the dataset to use for the development set.
            test_split: Proportion of the dataset to use for the test set.
            seed: Seed for reproducibility.
            metrics: List of metrics to calculate.
            tracking_metric: Metric to track.
            train_transform: Transform to apply to the training set.
                Defaults to None.
            dev_transform: Transform to apply to the development set.
                Defaults to None.
            test_transform: Transform to apply to the test set.
                Defaults to None.
        """
        if task not in TrainingConstants().TASKS:
            raise ValueError(
                f"Invalid task '{task}', "
                f"must be in {TrainingConstants().TASKS}."
            )
        self.size = size
        self.num_targets = num_targets
        self.feature_shape = feature_shape
        self._assert_splits(dev_split, test_split)
        self.dev_split = dev_split
        self.test_split = test_split
        self._generator = torch.Generator().manual_seed(seed)
        if dtype not in ["float32", "uint8"]:
            raise ValueError(
                f"Invalid dtype='{dtype}', must be in ['float32', 'uint8']."
            )
        self.dtype = dtype
        super().__init__(
            task=task,
            path="/",
            features_subdir="",
            file_type="",
            index_column="index",
            target_column="",
            file_handler="autrainer.datasets.utils.IdentityFileHandler",
            seed=seed,
            metrics=metrics,
            tracking_metric=tracking_metric,
            train_transform=train_transform,
            dev_transform=dev_transform,
            test_transform=test_transform,
            stratify=None,
        )
        self._mock_df  # required due to super().__init__ with placeholders

    @staticmethod
    def _assert_splits(dev_split: int, test_split: float) -> None:
        if dev_split + test_split >= 1.0:
            raise ValueError(
                f"Sum of dev_split '{dev_split}' and test_split "
                f"'{test_split}' must be less than 1."
            )

    @cached_property
    def _dataset_sizes(self) -> Tuple[int, int, int]:
        if self.size <= 0:
            raise ValueError(f"Size must be > 0, got '{self.size}'.")

        dev_size = int(self.size * self.dev_split)
        test_size = int(self.size * self.test_split)
        train_size = self.size - dev_size - test_size
        if any(size <= 0 for size in (train_size, dev_size, test_size)):
            raise ValueError(
                f"All subsets must be > 0, got train_size '{train_size}', "
                f"dev_size '{dev_size}', test_size '{test_size}'."
            )
        return train_size, dev_size, test_size

    @cached_property
    def _mock_df(self) -> pd.DataFrame:
        rng = np.random.default_rng(self.seed)

        if self.task == "ml-classification":
            self.target_column = [
                f"class_{i+1}" for i in range(self.num_targets)
            ]
            df = pd.DataFrame(
                {
                    **{
                        col: rng.integers(0, 2, size=self.size).astype(
                            np.float32
                        )
                        for col in self.target_column
                    }
                }
            )
        elif self.task == "classification":
            self.target_column = "target"
            df = pd.DataFrame(
                {
                    self.target_column: rng.integers(
                        0, self.num_targets, size=self.size
                    )
                }
            )
            df[self.target_column] = df[self.target_column].apply(
                lambda x: f"class_{x+1}"
            )
        elif self.task == "mt-regression":
            self.target_column = [
                f"target_{i+1}" for i in range(self.num_targets)
            ]
            df = pd.DataFrame(
                {
                    **{
                        col: rng.random((self.size,)).astype(np.float32)
                        for col in self.target_column
                    }
                }
            )
        elif self.task == "regression":
            self.target_column = "target"
            df = pd.DataFrame({self.target_column: rng.random((self.size,))})
        else:
            raise ValueError(
                f"Invalid task '{self.task}', "
                f"must be in {TrainingConstants().TASKS}."
            )

        return df

    @cached_property
    def df_train(self) -> pd.DataFrame:
        train_size, *_ = self._dataset_sizes
        return self._reset_index(self._mock_df.iloc[:train_size])

    @cached_property
    def df_dev(self) -> pd.DataFrame:
        train_size, dev_size, _ = self._dataset_sizes
        return self._reset_index(
            self._mock_df.iloc[train_size : train_size + dev_size]
        )

    @cached_property
    def df_test(self) -> pd.DataFrame:
        train_size, dev_size, _ = self._dataset_sizes
        return self._reset_index(self._mock_df.iloc[train_size + dev_size :])

    def _reset_index(self, df: pd.DataFrame) -> pd.DataFrame:
        return (
            df.copy()
            .reset_index(drop=True)
            .reset_index(names=[self.index_column])
        )

    def _init_dataset(
        self,
        df: pd.DataFrame,
        transform: SmartCompose,
    ) -> ToyDatasetWrapper:
        return ToyDatasetWrapper(
            df=df,
            target_column=self.target_column,
            feature_shape=self.feature_shape,
            dtype=self.dtype,
            generator=self._generator,
            transform=transform,
            target_transform=self.target_transform,
        )

    @cached_property
    def target_transform(self) -> AbstractTargetTransform:
        if self.task == "ml-classification":
            return MultiLabelEncoder(0.5, self.target_column)
        elif self.task == "classification":
            return LabelEncoder(
                self.df_train[self.target_column].unique().tolist()
            )
        elif self.task == "mt-regression":
            return MultiTargetMinMaxScaler(
                target=self.target_column,
                minimum=self.df_train[self.target_column].min().to_list(),
                maximum=self.df_train[self.target_column].max().to_list(),
            )
        elif self.task == "regression":
            return MinMaxScaler(
                target=self.target_column,
                minimum=self.df_train[self.target_column].min(),
                maximum=self.df_train[self.target_column].max(),
            )
        else:
            raise ValueError(
                f"Invalid task '{self.task}', "
                f"must be in {TrainingConstants().TASKS}."
            )



from typing import Dict, List, Union

import torch

from .abstract_target_transform import AbstractTargetTransform


class MinMaxScaler(AbstractTargetTransform):
    def __init__(self, target: str, minimum: float, maximum: float) -> None:
        """Minimum-Maximum Scaler for regression targets.

        Args:
            target: Name of the target.
            minimum: Minimum value of all target values.
            maximum: Maximum value of all target values.

        Raises:
            ValueError: If minimum is not less than maximum.
        """
        if not minimum < maximum:
            raise ValueError(
                f"Minimum '{minimum}' must be less than maximum '{maximum}'."
            )
        self.target = target
        self.minimum = float(minimum)
        self.maximum = float(maximum)

    def __len__(self) -> int:
        """Get the number of unique targets.

        Returns:
            Number of unique targets.
        """
        return 1

    def encode(self, x: float) -> float:
        """Encode a target value by scaling it between the minimum and maximum.

        Args:
            x: Target value.

        Returns:
            Scaled target value.
        """
        return (x - self.minimum) / (self.maximum - self.minimum)

    def decode(self, x: float) -> float:
        """Decode a target value by reversing the scaling between the minimum
        and maximum. Inverse operation of encode.

        Args:
            x: Scaled target value.

        Returns:
            Unscaled target value.
        """
        return x * (self.maximum - self.minimum) + self.minimum

    def probabilities_training(self, x: torch.Tensor) -> torch.Tensor:
        """Get the encoded probabilities from a batch of model outputs
        during training by applying the sigmoid function.

        Args:
            x: Batch of model outputs.

        Returns:
            Encoded probabilities.
        """
        return torch.sigmoid(x)

    def probabilities_inference(self, x: torch.Tensor) -> torch.Tensor:
        """Get the encoded probabilities from a batch of model outputs by
        applying the sigmoid function.

        Args:
            x: Batch of model outputs.

        Returns:
            Encoded probabilities.
        """
        return torch.sigmoid(x)

    def predict_inference(self, x: torch.Tensor) -> Union[List[float], float]:
        """Get the encoded predictions from a batch of model output
        probabilities by returning the raw values.

        Args:
            x: Batch of model output probabilities.

        Returns:
            Encoded predictions.
        """
        return x.squeeze().tolist()

    def majority_vote(self, x: List[float]) -> float:
        """Get the majority vote from a list of target values by averaging
        the predictions.

        Args:
            x: List of target values.

        Returns:
            Average target value.
        """
        return sum(x) / len(x)

    def probabilities_to_dict(self, x: torch.Tensor) -> Dict[str, float]:
        """Convert a tensor of probabilities to a dictionary of targets and
        their probabilities.

        Args:
            x: Tensor of probabilities.

        Returns:
            Dictionary of targets and their probabilities.
        """
        return {self.target: x.item()}



from typing import Dict, List, Union

import torch

from .abstract_target_transform import AbstractTargetTransform


class LabelEncoder(AbstractTargetTransform):
    def __init__(self, labels: List[str]) -> None:
        """Label encoder for single-label classification targets.

        Args:
            labels: List of target labels.
        """
        self.labels = sorted(labels)
        codes = range(len(self.labels))
        self.inverse_map = {
            code: label for code, label in zip(codes, self.labels)
        }
        self.map = {label: code for code, label in zip(codes, self.labels)}

    def __len__(self) -> int:
        """Get the number of unique target labels.

        Returns:
            Number of unique target labels.
        """
        return len(self.labels)

    def encode(self, x: str) -> int:
        """Encode a target label by mapping it to an integer.

        Args:
            x: Target label.

        Returns:
            Encoded target label.
        """
        return self.map[x]

    def decode(self, x: int) -> str:
        """Decode an encoded target label integer by mapping it to a label.

        Args:
            x: Encoded target label.

        Returns:
            Decoded target label.
        """
        return self.inverse_map[x]

    def probabilities_training(self, x: torch.Tensor) -> torch.Tensor:
        """Get the encoded probabilities from a batch of model outputs
        during training by returning the raw model outputs.

        Args:
            x: Batch of model outputs.

        Returns:
            Encoded probabilities.
        """
        return x

    def probabilities_inference(self, x: torch.Tensor) -> torch.Tensor:
        """Get the encoded probabilities from a batch of model outputs by
        applying the softmax function.

        Args:
            x: Batch of model outputs.

        Returns:
            Encoded probabilities.
        """
        return torch.softmax(x, dim=-1)

    def predict_inference(self, x: torch.Tensor) -> Union[List[int], int]:
        """Get the encoded predictions from a batch of model output
        probabilities by obtaining the index of the maximum value.

        Args:
            x: Batch of model output probabilities.

        Returns:
            Encoded predictions.
        """
        return x.argmax(-1).squeeze().tolist()

    def majority_vote(self, x: List[str]) -> str:
        """Get the majority vote from a list of decoded labels by
        determining the most frequently predicted label.

        Args:
            x: List of decoded labels.

        Returns:
            Decoded majority vote.
        """
        return max(set(x), key=x.count)

    def probabilities_to_dict(self, x: torch.Tensor) -> Dict[str, float]:
        """Convert a tensor of probabilities to a dictionary of labels and
        their probabilities.

        Args:
            x: Tensor of probabilities.

        Returns:
            Dictionary of labels and their probabilities.
        """
        return {self.decode(i): prob.item() for i, prob in enumerate(x)}



from functools import cached_property
import os
import shutil
from typing import Dict, List, Optional, Union

from omegaconf import DictConfig
import pandas as pd

from autrainer.transforms import SmartCompose

from .abstract_dataset import BaseClassificationDataset
from .utils import ZipDownloadManager


FILES = {"download.zip": "http://emodb.bilderbar.info/download/download.zip"}

EMOTIONS = {
    "W": "anger",
    "L": "boredom",
    "E": "disgust",
    "A": "fear",
    "F": "happiness",
    "T": "sadness",
    "N": "neutral",
}


class EmoDB(BaseClassificationDataset):
    def __init__(
        self,
        path: str,
        features_subdir: Optional[str],
        seed: int,
        metrics: List[Union[str, DictConfig, Dict]],
        tracking_metric: Union[str, DictConfig, Dict],
        index_column: str,
        target_column: str,
        file_type: str,
        file_handler: Union[str, DictConfig, Dict],
        features_path: Optional[str] = None,
        train_transform: Optional[SmartCompose] = None,
        dev_transform: Optional[SmartCompose] = None,
        test_transform: Optional[SmartCompose] = None,
        stratify: Optional[List[str]] = None,
        train_speakers: Optional[List[int]] = None,
        dev_speakers: Optional[List[int]] = None,
        test_speakers: Optional[List[int]] = None,
    ) -> None:
        """EmoDB dataset for the task of Speech Emotion Recognition.

        Args:
            path: Root path to the dataset.
            features_subdir: Subdirectory containing the features.
                If `None`, defaults to audio subdirectory,
                which is `default` for the standard format,
                but can be overridden in the dataset specification.
            seed: Seed for reproducibility.
            metrics: List of metrics to calculate.
            tracking_metric: Metric to track.
            index_column: Index column of the dataframe.
            target_column: Target column of the dataframe.
            file_type: File type of the features.
            file_handler: File handler to load the data.
            features_path: Root path to features. Useful
                when features need to be extracted and stored
                in a different directory than the root of the dataset.
                If `None`, will be set to `path`. Defaults to `None`.
            train_transform: Transform to apply to the training set.
                Defaults to None.
            dev_transform: Transform to apply to the development set.
                Defaults to None.
            test_transform: Transform to apply to the test set.
                Defaults to None.
            stratify: Columns to stratify the dataset on. Defaults to None.
            train_speakers: List of speakers IDs (int) to use for training.
                If None, 3, 8, 9, 10, 11, 12 are used. Defaults to None.
            dev_speakers: List of speakers IDs (int) to use for validation.
                If None, 13, 14 are used. Defaults to None.
            test_speakers: List of speakers IDs (int) to use for testing.
                If None, 15, 16 are used. Defaults to None.
        """
        self.train_speakers = train_speakers or [3, 8, 9, 10, 11, 12]
        self.dev_speakers = dev_speakers or [13, 14]
        self.test_speakers = test_speakers or [15, 16]
        super().__init__(
            path=path,
            features_subdir=features_subdir,
            seed=seed,
            metrics=metrics,
            tracking_metric=tracking_metric,
            index_column=index_column,
            target_column=target_column,
            file_type=file_type,
            file_handler=file_handler,
            features_path=features_path,
            train_transform=train_transform,
            dev_transform=dev_transform,
            test_transform=test_transform,
            stratify=stratify,
        )

    @cached_property
    def meta(self) -> pd.DataFrame:
        return pd.read_csv(os.path.join(self.path, "metadata.csv"))

    @cached_property
    def df_train(self) -> pd.DataFrame:
        return self.meta[self.meta["speaker"].isin(self.train_speakers)]

    @cached_property
    def df_dev(self) -> pd.DataFrame:
        return self.meta[self.meta["speaker"].isin(self.dev_speakers)]

    @cached_property
    def df_test(self) -> pd.DataFrame:
        return self.meta[self.meta["speaker"].isin(self.test_speakers)]

    @staticmethod
    def download(path: str) -> None:  # pragma: no cover
        """Download the EmoDB dataset.

        For more information on the dataset, see:
        http://emodb.bilderbar.info/docu/

        Args:
            path: Path to the directory to download the dataset to.
        """

        out_path = os.path.join(path, "default")
        if os.path.isdir(out_path):
            return

        dl_manager = ZipDownloadManager(FILES, path)
        dl_manager.download(check_exist=["download.zip"])
        dl_manager.extract(check_exist=["wav"])
        shutil.move(os.path.join(path, "wav"), out_path)
        files = os.listdir(out_path)
        files = sorted(files)

        df = pd.DataFrame(files, columns=["filename"])
        df["emotion"] = df["filename"].str[5].map(EMOTIONS)
        df["speaker"] = df["filename"].str[0:2]
        df["text"] = df["filename"].str[2:5]
        df["version"] = df["filename"].str[6]

        df.to_csv(os.path.join(path, "metadata.csv"), index=False)

        # remove unnecessary files
        shutil.rmtree(os.path.join(path, "lablaut"))
        shutil.rmtree(os.path.join(path, "labsilb"))
        shutil.rmtree(os.path.join(path, "silb"))
        os.remove(os.path.join(path, "erkennung.txt"))
        os.remove(os.path.join(path, "erklaerung.txt"))



from typing import Optional, Tuple, Union
import warnings

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from .abstract_model import AbstractModel
from .timm_wrapper import TimmModel


with warnings.catch_warnings():
    warnings.simplefilter("ignore", category=FutureWarning)
    from speechbrain.lobes.features import Leaf as LeafSb


class LEAFNet(AbstractModel):
    def __init__(
        self,
        output_dim: int,
        leaf_filters: int = 40,
        kernel_size: int = 25,
        stride: float = 0.0625,
        window_stride: int = 10,
        padding_kernel_size: int = 25,
        sample_rate: int = 16000,
        min_freq: int = 60,
        max_freq: int = 7800,
        efficientnet_type: str = "efficientnet_b0",
        mode: str = "interspeech",
        initialization: str = "mel",
        generator_seed: int = 42,
        transfer: bool = False,
    ) -> None:
        """EfficientNet with LEAF-Is frontend.
        Used to reproduce work from:
        https://www.isca-archive.org/interspeech_2023/meng23c_interspeech.html

        Also see original LEAF and PCEN papers (c.f. speechbrain).

        We take and slightly adapt the LEAF frontend implementation from:
        https://github.com/Hanyu-Meng/Adapting-LEAF

        Args:
            output_dim: Output dimension.
            leaf_filters: Number of LEAF filterbanks to train. Defaults to 40.
            kernel_size: Size of kernels applied by LEAF (in ms).
                Defaults to 25.
            stride: Stride of LEAF (in ms). Defaults to 0.0625.
            window_stride: Stride of lowpass filter (in ms). Defaults to 10.
            padding_kernel_size: Size of lowpass filter (in ms).
                Defaults to 25.
            sample_rate: Used to compute LEAF params. Defaults to 16000.
            min_freq: Minimum freq analyzed by LEAF. Defaults to 60.
            max_freq: Maximum freq analyzed by LEAF. Defaults to 7800.
            efficientnet_type: EfficientNet type to use from timm.
                Defaults to "efficientnet_b0".
            mode: Implementation according to "interspeech" paper (Meng et al.)
                or "speech_brain". Defaults to "interspeech".
            initialization: Filterbank initialisation in ["mel", "bark",
                "linear-constant", "constant", "uniform", "zeros"].
                Defaults to "mel".
            generator_seed: Seed for random generator. Defaults to 42.
            transfer: Whether to use EfficientNet weights from ImageNet.
                Defaults to False.

        Raises:
            ValueError: If efficientnet_type is not supported.
            ValueError: If mode is not supported.
        """
        super().__init__(output_dim, transfer)
        # convert LEAF params from ms to samples
        self.mode = mode
        self.initialization = initialization
        self.leaf_filters = leaf_filters
        self.min_freq = min_freq
        self.max_freq = max_freq
        self.sample_rate = sample_rate
        self.generator_seed = generator_seed
        self.kernel_size = kernel_size
        self.stride = stride
        self.window_stride = window_stride
        self.padding_kernel_size = padding_kernel_size
        self.efficientnet_type = efficientnet_type
        kernel_size_sample = int(sample_rate * kernel_size / 1000)
        kernel_size_sample += 1 - (kernel_size % 2)  # make odd

        # convert pooling params from ms to samples
        if mode == "interspeech":
            self.leaf = LeafIs(
                n_filters=leaf_filters,
                min_freq=min_freq,
                max_freq=max_freq,
                sample_rate=sample_rate,
                window_len=padding_kernel_size,
                window_stride=window_stride,
            )

        elif mode == "speech_brain":
            self.leaf = LeafSb(
                out_channels=leaf_filters,
                in_channels=1,
                sample_rate=sample_rate,
                min_freq=min_freq,
                use_pcen=True,
                learnable_pcen=True,
                use_legacy_complex=True,
                skip_transpose=True,
            )
        else:
            raise ValueError(
                "Only options 'interspeech' and 'speech_brain' are available"
                f" for mode, but got mode='{mode}'."
            )
        self._initialise_filterbank()
        if not self.efficientnet_type.startswith("efficientnet_"):
            raise ValueError(
                "Only EfficientNet models are supported, but got"
                f" efficientnet_type='{efficientnet_type}'."
            )
        self.classifier = TimmModel(
            output_dim=self.output_dim,
            timm_name=self.efficientnet_type,
            transfer=self.transfer,
            in_chans=1,
        )

    def embeddings(self, features: torch.Tensor) -> torch.Tensor:
        return self.leaf(features)

    def forward(self, features: torch.Tensor) -> torch.Tensor:
        x = self.leaf(features)
        x = x.unsqueeze(1)
        x = self.classifier(x)
        return x

    def _initialise_filterbank(self) -> None:
        if self.initialization == "mel":
            return
        elif self.initialization == "linear-constant":
            center_frequencies = torch.linspace(
                self.min_freq, self.max_freq, self.leaf_filters
            )
            bandwidths = torch.ones((self.leaf_filters,)) * np.exp(
                (np.log(self.max_freq) - np.log(self.min_freq)) / 2.5
                + np.log(self.min_freq)
            )
        elif self.initialization == "zeros":
            center_frequencies = torch.zeros((self.leaf_filters,))
            bandwidths = torch.ones((self.leaf_filters,)) * np.exp(
                np.log(self.max_freq - self.min_freq) / 2
            )
        elif self.initialization == "constant":
            center_frequencies = torch.ones((self.leaf_filters,)) * np.exp(
                (np.log(self.max_freq) - np.log(self.min_freq)) / 2
                + np.log(self.min_freq)
            )
            bandwidths = torch.ones((self.leaf_filters,)) * np.exp(
                (np.log(self.max_freq) - np.log(self.min_freq)) / 2.5
                + np.log(self.min_freq)
            )
        elif self.initialization == "uniform":
            generator = torch.Generator()
            generator.manual_seed(self.generator_seed)
            center_frequencies = (
                torch.rand((self.leaf_filters,), generator=generator)
                * (self.max_freq - self.min_freq)
                + self.min_freq
            )
            center_frequencies = torch.sort(center_frequencies).values
            # Estimation
            bandwidths = (
                torch.rand((self.leaf_filters,), generator=generator)
                * (self.max_freq - self.min_freq)
                / 10
                + self.min_freq
            )
        elif self.initialization == "bark":
            center_frequencies, bandwidths = bark_scale_filterbank(
                self.min_freq, self.max_freq, self.leaf_filters
            )
        else:
            raise ValueError(
                "Only options 'mel', 'linear-constant', 'constant', 'uniform', "
                f"'bark', and 'zeros' are available for initialization "
                f", but got initialization='{self.initialization}'."
            )

        # adjustment for sample_rate
        center_frequencies *= 2 * np.pi / self.sample_rate
        # bandwiths to sigmas
        bandwidths = (self.sample_rate / 2.0) / bandwidths

        if self.mode == "interspeech":
            center_freqs_param = "filterbank.center_freqs"
            bandwith_param = "filterbank.bandwidths"
            self.leaf.state_dict()[center_freqs_param].copy_(
                center_frequencies
            )
            self.leaf.state_dict()[bandwith_param].copy_(bandwidths)
        elif self.mode == "speech_brain":
            filterbank_param = "complex_conv.kernel"
            self.leaf.state_dict()[filterbank_param][:, 0].copy_(
                center_frequencies
            )
            self.leaf.state_dict()[filterbank_param][:, 1].copy_(bandwidths)


def hz_to_bark(
    f: Union[int, float, np.ndarray],
) -> Union[int, float, np.ndarray]:
    """Convert frequency from Hz to Bark scale according to Traunmüller.

    Args:
        f: Frequency in Hz.

    Returns:
        Frequency in Bark.
    """
    b = 26.81 * f / (1960 + f) - 0.53
    return b


def bark_to_hz(
    b: Union[int, float, np.ndarray],
) -> Union[int, float, np.ndarray]:
    """Approximate conversion from Bark to Hz (inverse of the above).
    Ajdusted centers for reconversion according to Traunmüller.

    Args:
        b: Frequency in Bark.

    Returns:
        Frequency in Hz.
    """

    z_1 = b[b < 2] + 0.15 * (2 - b[b < 2])
    z_2 = b[b >= 2]
    z_2 = z_2[z_2 <= 20.1]
    z_3 = b[b > 20.1] + 0.22 * (b[b > 20.1] - 20.1)
    z = np.hstack((z_1, z_2, z_3))
    f = 1960 * (z + 0.53) / (26.28 - z)
    return f  # 600 * np.sinh(z / 6)


def bark_scale_filterbank(
    min_freq: Union[int, float],
    max_freq: Union[int, float],
    num_bands: int,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Calculate center frequencies and bandwidths for a Bark scale filterbank.

    Args:
        min_freq: Minimum frequency in Hz.
        max_freq: Maximum frequency in Hz.
        num_bands: Number of bands.

    Returns:
        Tuple of center frequencies and bandwidths.
    """
    # Convert min and max frequencies to Bark scale
    min_bark = hz_to_bark(min_freq)
    max_bark = hz_to_bark(max_freq)

    # Evenly distribute bands in Bark scale
    bark_centers = np.linspace(
        min_bark, max_bark + (max_bark - min_bark) / num_bands, num_bands + 1
    )

    # Convert center frequencies back to Hz
    center_freqs = bark_to_hz(bark_centers)
    bandwidths = np.diff(center_freqs)

    return torch.Tensor(center_freqs[:-1]), torch.Tensor(bandwidths)


def mel_filter_params(
    n_filters: int,
    min_freq: float,
    max_freq: float,
    sample_rate: int,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Analytically calculate the center frequencies and sigmas of a mel
    filter bank.

    Args:
        n_filters: Number of filters for the filterbank.
        min_freq: Minimum cutoff for the frequencies.
        max_freq: Maximum cutoff for the frequencies.
        sample_rate: Sample rate to use for the calculation.

    Returns:
        Center frequencies and sigmas.
    """
    min_mel = 1127 * np.log1p(min_freq / 700.0)
    max_mel = 1127 * np.log1p(max_freq / 700.0)
    peaks_mel = torch.linspace(min_mel, max_mel, n_filters + 2)
    peaks_hz = 700 * (torch.expm1(peaks_mel / 1127))
    center_freqs = peaks_hz[1:-1] * (2 * np.pi / sample_rate)
    bandwidths = peaks_hz[2:] - peaks_hz[:-2]
    sigmas = (sample_rate / 2.0) / bandwidths
    return center_freqs, sigmas


def gabor_filters(
    size: int, center_freqs: torch.Tensor, sigmas: torch.Tensor
) -> torch.Tensor:
    """Calculate a gabor function from given center frequencies and bandwidths
    that can be used as kernel/filters for an 1D convolution.

    Args:
        size: Kernel/filter size.
        center_freqs: Center frequencies.
        sigmas: Bandwidths.

    Returns:
        Kernel/filter that can be used for an 1D convolution.
    """
    t = torch.arange(-(size // 2), (size + 1) // 2, device=center_freqs.device)
    denominator = 1.0 / (np.sqrt(2 * np.pi) * sigmas)
    gaussian = torch.exp(torch.outer(1.0 / (2.0 * sigmas**2), -(t**2)))
    sinusoid = torch.exp(1j * torch.outer(center_freqs, t))
    return denominator[:, np.newaxis] * sinusoid * gaussian


def gauss_windows(size: int, sigmas: torch.Tensor) -> torch.Tensor:
    """Calculate a gaussian lowpass function from given bandwidths that can be
    used as kernel/filter for an 1D convolution.

    Args:
        size: Kernel/filter size.
        sigmas: Bandwidths.

    Returns:
        Kernel/filter that can be used for an 1D convolution.
    """
    t = torch.arange(0, size, device=sigmas.device)
    numerator = t * (2 / (size - 1)) - 1
    return torch.exp(-0.5 * (numerator / sigmas[:, np.newaxis]) ** 2)


class GaborFilterbank(nn.Module):
    def __init__(
        self,
        n_filters: int,
        min_freq: float,
        max_freq: float,
        sample_rate: int,
        filter_size: int,
        pool_size: int,
        pool_stride: int,
        pool_init: float = 0.4,
    ) -> None:
        """Torch module that functions as a gabor filterbank.

        Initializes n_filters center frequencies and bandwidths that are based
        on a mel filterbank. The parameters are used to calculate Gabor filters
        for a 1D convolution over the input signal. The squared modulus is
        taken from the results. To reduce the temporal resolution a gaussian
        lowpass filter is calculated from pooling_widths, which are used to
        perform a pooling operation. The center frequencies, bandwidths and
        pooling_widths are learnable parameters.

        Args:
            n_filters: Number of filters.
            min_freq: Minimum frequency for the mel filterbank initialization.
            max_freq: Maximum frequency for the mel filterbank initialization.
            sample_rate: Sample rate for the mel filterbank initialization.
            filter_size: Size of the kernels/filters for gabor convolution.
            pool_size: Size of the kernels/filters for pooling convolution.
            pool_stride: Stride of the pooling convolution.
            pool_init: Initial value for the gaussian lowpass function.
                Defaults to 0.4.
        """
        super(GaborFilterbank, self).__init__()
        self.n_filters = n_filters
        self.filter_size = filter_size
        self.pool_size = pool_size
        self.pool_stride = pool_stride
        center_freqs, bandwidths = mel_filter_params(
            n_filters, min_freq, max_freq, sample_rate
        )
        self.center_freqs = nn.Parameter(center_freqs)
        self.bandwidths = nn.Parameter(bandwidths)
        self.pooling_widths = nn.Parameter(
            torch.full((n_filters,), float(pool_init))
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # compute filters
        center_freqs = self.center_freqs.clamp(min=0.0, max=np.pi)
        z = np.sqrt(2 * np.log(2)) / np.pi
        bandwidths = self.bandwidths.clamp(min=4 * z, max=self.filter_size * z)
        filters = gabor_filters(self.filter_size, center_freqs, bandwidths)
        filters = torch.cat((filters.real, filters.imag), dim=0).unsqueeze(1)
        # convolve with filters
        x = F.conv1d(x, filters, padding=self.filter_size // 2)
        # compute squared modulus
        x = x**2
        x = x[:, : self.n_filters] + x[:, self.n_filters :]
        # compute pooling windows
        pooling_widths = self.pooling_widths.clamp(
            min=2.0 / self.pool_size, max=0.5
        )
        windows = gauss_windows(self.pool_size, pooling_widths).unsqueeze(1)
        # apply temporal pooling
        x = F.conv1d(
            x,
            windows,
            stride=self.pool_stride,
            padding=self.filter_size // 2,
            groups=self.n_filters,
        )
        return x


class PCEN(nn.Module):
    def __init__(
        self,
        num_bands: int,
        s: float = 0.025,
        alpha: float = 1.0,
        delta: float = 1.0,
        r: float = 1.0,
        eps: float = 1e-6,
        learn_logs: bool = False,
        clamp: Optional[float] = None,
    ) -> None:
        """Trainable PCEN (Per-Channel Energy Normalization) layer.

        .. math::
            Y = (\\frac{X}{(\\epsilon + M)^\\alpha} + \\delta)^r - \\delta^r
            M_t = (1 - s) M_{t - 1} + s X_t

        Args:
            num_bands: Number of frequency bands (before last input dimension).
            s: Initial value for :math:`s`.
            alpha: Initial value for :math:`alpha`
            delta: Initial value for :math:`delta`
            r: Initial value for :math:`r`
            eps: Value for :math:`eps`
            learn_logs: If false-ish, instead of learning the logarithm of each
                parameter (as in the PCEN paper), learn the inverse of :math:`r`
                and all other parameters directly (as in the LEAF paper).
            clamp: If given, clamps the input to the given minimum value before
                applying PCEN.
        """
        super(PCEN, self).__init__()
        if learn_logs:
            # learns logarithm of each parameter
            s = np.log(s)
            alpha = np.log(alpha)
            delta = np.log(delta)
            r = np.log(r)
        else:
            # learns inverse of r, and all other parameters directly
            r = 1.0 / r
        self.learn_logs = learn_logs
        self.s = nn.Parameter(torch.full((num_bands,), float(s)))
        self.alpha = nn.Parameter(torch.full((num_bands,), float(alpha)))
        self.delta = nn.Parameter(torch.full((num_bands,), float(delta)))
        self.r = nn.Parameter(torch.full((num_bands,), float(r)))
        self.eps = torch.as_tensor(eps)
        self.clamp = clamp

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # clamp if needed
        if self.clamp is not None:
            x = x.clamp(min=self.clamp)

        # prepare parameters
        if self.learn_logs:
            # learns logarithm of each parameter
            s = self.s.exp()
            alpha = self.alpha.exp()
            delta = self.delta.exp()
            r = self.r.exp()
        else:
            # learns inverse of r, and all other parameters directly
            s = self.s
            alpha = self.alpha.clamp(max=1)
            delta = self.delta.clamp(min=0)  # unclamped in original LEAF impl.
            r = 1.0 / self.r.clamp(min=1)
        # broadcast over channel dimension
        alpha = alpha[:, np.newaxis]
        delta = delta[:, np.newaxis]
        r = r[:, np.newaxis]

        # compute smoother
        smoother = [x[..., 0]]  # initialize the smoother with the first frame
        for frame in range(1, x.shape[-1]):
            smoother.append((1 - s) * smoother[-1] + s * x[..., frame])
        smoother = torch.stack(smoother, -1)

        # stable reformulation due to Vincent Lostanlen; original formula was:
        # return (input / (self.eps + smoother)**alpha + delta)**r - delta**r
        smoother = torch.exp(
            -alpha * (torch.log(self.eps) + torch.log1p(smoother / self.eps))
        )
        return (x * smoother + delta) ** r - delta**r


class LeafIs(nn.Module):
    def __init__(
        self,
        n_filters: int = 40,
        min_freq: float = 60.0,
        max_freq: float = 7800.0,
        sample_rate: int = 16000,
        window_len: float = 25.0,
        window_stride: float = 10.0,
        compression: Optional[torch.nn.Module] = None,
    ) -> None:
        """LEAF frontend, a learnable front-end that takes an audio waveform as
        input and outputs a learnable spectral representation.
        Initially approximates the computation of standard mel-filterbanks.

        Args:
            n_filters: Number of filters. Defaults to 40.
            min_freq: Minimum frequency. Defaults to 60.0.
            max_freq: Maximum frequency. Defaults to 7800.0.
            sample_rate: Sample Rate for filterbanc initialization.
                Defaults to 16000.
            window_len: Kernel/filter size of the convolutions in ms.
                Defaults to 25.0.
            window_stride: Stride used for the pooling convolution in ms.
                Defaults to 10.0.
            compression: Compression function. If None, PCEN is used.
                Defaults to None.
        """
        super(LeafIs, self).__init__()

        # convert window sizes from milliseconds to samples
        window_size = int(sample_rate * window_len / 1000)
        window_size += 1 - (window_size % 2)  # make odd
        window_stride = int(sample_rate * window_stride / 1000)

        self.filterbank = GaborFilterbank(
            n_filters,
            min_freq,
            max_freq,
            sample_rate,
            filter_size=window_size,
            pool_size=window_size,
            pool_stride=window_stride,
        )

        self.compression = (
            compression
            if compression
            else PCEN(
                n_filters,
                s=0.04,
                alpha=0.96,
                delta=2,
                r=0.5,
                eps=1e-12,
                learn_logs=False,
                clamp=1e-5,
            )
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        while x.ndim < 3:
            x = x[:, np.newaxis]
        x = self.filterbank(x)
        x = self.compression(x)
        return x
